{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       Franois PIETTE @ OverByte
Creation:     Sep 26, 2012
Description:
Version:      1.00
History:


 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit AdoDBSchemaDemoMain;

interface

uses
    Windows, Messages, SysUtils, Variants, Classes, Graphics,
    Controls, Forms, Dialogs, ShlObj, IniFiles, Vcl.StdCtrls,
    TypInfo,
    Data.DB,
    Data.Win.ADODB, ADOInt, Vcl.ComCtrls;

const
    WM_APP_STARTUP       = WM_USER + 1;
    CompanyFolder        = 'OverByte';

type
    TDBCOLUMNFLAGSENUM = (
        DBCOLUMNFLAGS_ISBOOKMARK        = $00001,
        DBCOLUMNFLAGS_MAYDEFER          = $00002,
        DBCOLUMNFLAGS_WRITE             = $00004,
        DBCOLUMNFLAGS_WRITEUNKNOWN      = $00008,
        DBCOLUMNFLAGS_ISFIXEDLENGTH     = $00010,
        DBCOLUMNFLAGS_ISNULLABLE        = $00020,
        DBCOLUMNFLAGS_MAYBENULL         = $00040,
        DBCOLUMNFLAGS_ISLONG            = $00080,
        DBCOLUMNFLAGS_ISROWID           = $00100,
        DBCOLUMNFLAGS_ISROWVER          = $00200,
        DBCOLUMNFLAGS_CACHEDEFERRED     = $00400,
        DBCOLUMNFLAGS_SCALEISNEGATIVE   = $00800,
        DBCOLUMNFLAGS_RESERVED          = $01000,
        DBCOLUMNFLAGS_ISROWURL          = $02000,
        DBCOLUMNFLAGS_ISDEFAULTSTREAM   = $04000,
        DBCOLUMNFLAGS_ISCOLLECTION      = $08000,
        DBCOLUMNFLAGS_ISSTREAM          = $10000,
        DBCOLUMNFLAGS_ISROWSET          = $20000,
        DBCOLUMNFLAGS_ISROW             = $40000,
        DBCOLUMNFLAGS_ROWSPECIFICCOLUMN = $80000);

    TNodeObject = class
        Name : String;
        constructor Create(const AName : String); virtual;
    end;

    TTableNodeObject = class(TNodeObject)
        Initialized : Boolean;
    end;

    TFieldNodeObject = class(TNodeObject)
        DataType        : DataTypeEnum;
        Precision       : Integer;
        Scale           : Integer;
        FieldLen        : Integer;
        FieldType       : TFieldType;
        Flags           : DWORD;
        function ToString : String; override;
    end;

    TAdoSchemaForm = class(TForm)
        ADOConnection1: TADOConnection;
        LoadButton: TButton;
        TreeView1: TTreeView;
        procedure LoadButtonClick(Sender: TObject);
        procedure TreeView1Click(Sender: TObject);
        procedure TreeView1MouseMove(Sender: TObject; Shift: TShiftState; X,
          Y: Integer);
    protected
        FLocalAppData           : String;
        FAppName                : String;
        FIniFileName            : String;
        FInitialized            : Boolean;
        FIniSection             : String;
        FIniSectionData         : String;
        FParentedFlag           : Boolean;
        FTables                 : TStringList;
        FFields                 : TStringList;
        FMouseX                 : Integer;
        FMouseY                 : Integer;
        FOnStartup              : TNotifyEvent;
        procedure WMAppStartup(var Msg: TMessage); message WM_APP_STARTUP;
        procedure CreateParams(var Params: TCreateParams); override;
        procedure DoShow; override;
        procedure DoClose(var Action: TCloseAction); override;
        procedure DoStartup; virtual;
        procedure ClearTreeView(Tv: TTreeView);
        procedure ClearTreeNode(Node: TTreeNode);
        procedure GetTableNodeDetails(Node: TTreeNode);
    public
        constructor Create(AOwner : TComponent); override;
        constructor CreateParented(ParentWindow: HWnd); virtual;
        destructor Destroy; override;
        property IniFileName          : String       read  FIniFileName
                                                     write FIniFileName;
        property IniSection           : String       read  FIniSection
                                                     write FIniSection;
        property IniSectionData       : String       read  FIniSectionData
                                                     write FIniSectionData;
        property LocalAppData         : String       read  FLocalAppData
                                                     write FLocalAppData;
        property AppName              : String       read  FAppName
                                                     write FAppName;
        property OnStartup            : TNotifyEvent read  FOnStartup
                                                     write FOnStartup;
    end;

var
    AdoSchemaForm : TAdoSchemaForm;

implementation

{$R *.dfm}

const
    SectionWindow      = 'Window';   // Must be unique for each window
                                     // See IniSection property !
    SectionData        = 'Data';
    KeyTop             = 'Top';
    KeyLeft            = 'Left';
    KeyWidth           = 'Width';
    KeyHeight          = 'Height';

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

{ TAdoSchemaForm }

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
constructor TAdoSchemaForm.Create(AOwner: TComponent);
var
    Path : array [0..MAX_PATH] of Char;
begin
    SHGetFolderPath(0, CSIDL_LOCAL_APPDATA, 0, SHGFP_TYPE_CURRENT, @Path[0]);
    FIniSection     := SectionWindow;
    FIniSectionData := SectionData;
    FAppName        := ChangeFileExt(ExtractFileName(Application.ExeName), '');
    FLocalAppData   := IncludeTrailingPathDelimiter(Path) +
                       CompanyFolder + '\' + FAppName + '\';
    FIniFileName    := FLocalAppData + FAppName + '.ini';
    inherited Create(AOwner);
    if FParentedFlag then
        BorderStyle := bsNone;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
constructor TAdoSchemaForm.CreateParented(ParentWindow: HWnd);
begin
    FParentedFlag := TRUE;
    inherited CreateParented(ParentWindow);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.CreateParams(var Params: TCreateParams);
begin
    inherited CreateParams(Params);
    if not FParentedFlag then begin
        // This will make the form have an icon on taskbar and make it
        // separate from the main window.
        // This is not desirable if the window is created parented,
        // hence the flag.
        Params.ExStyle   := Params.ExStyle or WS_EX_APPWINDOW;
        Params.WndParent := GetDesktopwindow;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
destructor TAdoSchemaForm.Destroy;
begin
    ClearTreeView(TreeView1);
    inherited;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.DoClose(var Action: TCloseAction);
var
    IniFile : TIniFile;
begin
    try
        IniFile := TIniFile.Create(FIniFileName);
        try
            if not FParentedFlag then begin
                IniFile.WriteInteger(FIniSection, KeyTop,    Top);
                IniFile.WriteInteger(FIniSection, KeyLeft,   Left);
                IniFile.WriteInteger(FIniSection, KeyWidth,  Width);
                IniFile.WriteInteger(FIniSection, KeyHeight, Height);
            end;
        finally
            IniFile.Destroy;
        end;
    except
        // Ignore any exception when saving window size and position
    end;
    inherited DoClose(Action);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.DoShow;
var
    IniFile : TIniFile;
    I       : Integer;
    Ri      : TRect;
const
    Margin  = 100;
begin
    if not FInitialized then begin
        FInitialized := TRUE;

        ForceDirectories(ExtractFilePath(FIniFileName));
        IniFile := TIniFile.Create(FIniFileName);
        try
            if not FParentedFlag then begin
                Width  := IniFile.ReadInteger(FIniSection, KeyWidth,  Width);
                Height := IniFile.ReadInteger(FIniSection, KeyHeight, Height);
                Top    := IniFile.ReadInteger(FIniSection, KeyTop,
                                              (Screen.Height - Height) div 2);
                Left   := IniFile.ReadInteger(FIniSection, KeyLeft,
                                              (Screen.Width  - Width)  div 2);
            end;
        finally
            IniFile.Destroy;
        end;

        if not FParentedFlag then begin
            // Check if form is on an existing monitor
            I := 0;
            while I < Screen.MonitorCount do begin
                Windows.IntersectRect(Ri, BoundsRect, Screen.Monitors[I].BoundsRect);
                if (Ri.Width > Margin) and (Ri.Height > Margin) then
                    break;
                Inc(I);
            end;
            if I >= Screen.MonitorCount then begin
                // Form is outside of any monitor. Move to center of main monitor
                Top  := (Screen.Height - Height) div 2;
                Left := (Screen.Width  - Width)  div 2;
            end;
        end;

        PostMessage(Handle, WM_APP_STARTUP, 0, 0);
    end;
    inherited DoShow;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.DoStartup;
begin
    ADOConnection1.ConnectionString :=
        'Provider=SQLOLEDB.1;' +          // Driver to use
        'User ID=fpiette;' +              // User code
        'Password=mypassword;' +          // Password
        'Initial Catalog=Pointage;' +     // Database name
        'Data Source=ML150';              // Server name
    if Assigned(FOnStartup) then
        FOnStartup(Self);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.WMAppStartup(var Msg: TMessage);
begin
    DoStartup;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Remove all nodes from a TTreeView and free all objects associated with
// each node
procedure TAdoSchemaForm.ClearTreeView(Tv : TTreeView);
var
    I    : Integer;
    Obj  : TObject;
begin
    for I := 0 to Tv.Items.Count - 1 do begin
        Obj := TObject(Tv.Items[I].Data);
        Obj.Free;
        Tv.Items[I].Data := nil;
    end;
    Tv.Items.Clear;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Remove a node and all of its subnode in a TTreeView, freeing the objects
// associated with each node
procedure TAdoSchemaForm.ClearTreeNode(Node : TTreeNode);
var
    I     : Integer;
    Obj   : TObject;
    Child : TTreeNode;
begin
    if not Assigned(Node) then
        Exit;
    Obj := TObject(Node.Data);
    Obj.Free;
    Node.Data := nil;
    for I := 0 to Node.Count - 1 do begin
        Child := Node.Item[I];
        ClearTreeNode(Child);
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.LoadButtonClick(Sender: TObject);
var
    TableName : String;
    TableNodeObject : TTableNodeObject;
    Node            : TTreeNode;
begin
    if not Assigned(FTables) then
        FTables := TStringList.Create
    else
        FTables.Clear;
    ADOConnection1.Connected := TRUE;
    ADOConnection1.GetTableNames(FTables, FALSE);
    ClearTreeView(TreeView1);
    for TableName in FTables do begin
        TableNodeObject := TTableNodeObject.Create(TableName);
        Node := TreeView1.Items.AddObject(nil, TableName, TableNodeObject);
        // Add a fake child so that the expand marker is shown
        TreeView1.Items.AddChildObject(Node, 'reading...', nil);
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

{ TNodeObject }

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
constructor TNodeObject.Create(const AName: String);
begin
    inherited Create;
    Name := AName;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function ADOTypeToStr(const ADOType: DataTypeEnum) : String;
begin
    case ADOType of
    adEmpty :            Result := 'Empty';
    adTinyInt :          Result := 'TinyInt';
    adSmallInt :         Result := 'SmallInt';
    adInteger :          Result := 'Integer';
    adBigInt :           Result := 'BigInt';
    adUnsignedTinyInt :  Result := 'UnsignedTinyInt';
    adUnsignedSmallInt : Result := 'UnsignedSmallInt';
    adUnsignedInt :      Result := 'UnsignedInt';
    adUnsignedBigInt :   Result := 'UnsignedBigInt';
    adSingle :           Result := 'Single';
    adDouble :           Result := 'Double';
    adCurrency :         Result := 'Currency';
    adDecimal :          Result := 'Decimal';
    adNumeric :          Result := 'Numeric';
    adBoolean :          Result := 'Boolean';
    adError :            Result := 'Error';
    adUserDefined :      Result := 'UserDefined';
    adVariant :          Result := 'Variant';
    adIDispatch :        Result := 'IDispatch';
    adIUnknown :         Result := 'IUnknown';
    adGUID :             Result := 'GUID';
    adDate :             Result := 'Date';
    adDBDate :           Result := 'DBDate';
    adDBTime :           Result := 'DBTime';
    adDBTimeStamp :      Result := 'DBTimeStamp';
    adBSTR :             Result := 'BSTR';
    adChar :             Result := 'Char';
    adVarChar :          Result := 'VarChar';
    adLongVarChar :      Result := 'LongVarChar';
    adWChar :            Result := 'WChar';
    adVarWChar :         Result := 'VarWChar';
    adLongVarWChar :     Result := 'LongVarWChar';
    adBinary :           Result := 'Binary';
    adVarBinary :        Result := 'VarBinary';
    adLongVarBinary :    Result := 'LongVarBinary';
    adChapter :          Result := 'Chapter';
    adFileTime :         Result := 'FileTime';
    adDBFileTime :       Result := 'DBFileTime';
    adPropVariant :      Result := 'PropVariant';
    adVarNumeric :       Result := 'VarNumeric';
    adArray :            Result := 'Array';
    else
                         Result := IntToStr(ADOType);
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// http://msdn.microsoft.com/en-us/library/ms175970.aspx
// Using OLEDB and SQL server, char and varchar both have ADOType = adChar
// but DBCOLUMNFLAGS_ISFIXEDLENGTH = false for varchar and true for char
function ADOTypeToFieldType(
    const ADOType: DataTypeEnum;
    Precision, Scale: Integer;
    EnableBCD: Boolean = True): TFieldType;
begin
    case ADOType of
    adEmpty: Result := ftUnknown;
    adTinyInt, adSmallInt: Result := ftSmallint;
    adError, adInteger, adUnsignedInt: Result := ftInteger;
    adBigInt, adUnsignedBigInt: Result := ftLargeInt;
    adUnsignedTinyInt, adUnsignedSmallInt: Result := ftWord;
    adSingle, adDouble: Result := ftFloat;
    adCurrency: Result := ftBCD;
    adBoolean: Result := ftBoolean;
    adDBDate: Result := ftDate;
    adDBTime: Result := ftTime;
    adDate, adDBTimeStamp, adFileTime, adDBFileTime: Result := ftDateTime;
    adChar: Result := ftFixedChar;
    adVarChar: Result := ftString;
    adWChar: Result := ftFixedWideChar;
    adBSTR, adVarWChar: Result := ftWideString;
    adLongVarChar: Result := ftMemo;
    adLongVarWChar: Result := ftWideMemo;
    adLongVarBinary: Result := ftBlob;
    adBinary: Result := ftBytes;
    adVarBinary: Result := ftVarBytes;
    adChapter: Result := ftDataSet;
    adPropVariant, adVariant: Result := ftVariant;
    adIUnknown: Result := ftInterface;
    adIDispatch: Result := ftIDispatch;
    adGUID: Result := ftGUID;
    adDecimal, adNumeric, adVarNumeric:
        if EnableBCD then begin
          // Use TFMTBcd for precision and numeric scale values that are out of
          // the range of what is supported by TBCDField and the Currency type.
          if (Scale > 4) or (Precision > 19) then
              Result := ftFMTBcd
          else
              Result := ftBcd;
        end
        else
            Result := ftFloat;
    else
        Result := ftUnknown;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.TreeView1Click(Sender: TObject);
begin
    GetTableNodeDetails(TreeView1.GetNodeAt(FMouseX, FMouseY));
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.GetTableNodeDetails(
    Node : TTreeNode);
var
    TableNodeObject : TTableNodeObject;
    FieldNodeObject : TFieldNodeObject;
    FieldName       : String;
    Fields          : Recordset;
begin
    if Node = nil then
        Exit;
    if Node.Data = nil then
        Exit;

    if TObject(Node.Data) is TTableNodeObject then begin
        TableNodeObject := TObject(Node.Data) as TTableNodeObject;
        if not TableNodeObject.Initialized then begin
            TableNodeObject.Initialized := TRUE;
            ClearTreeNode(Node.Item[0]);
            Node.DeleteChildren;
            if not Assigned(FFields) then
                FFields := TStringList.Create
            else
                FFields.Clear;
            ADOConnection1.Connected := TRUE;
            Fields := ADOConnection1.ConnectionObject.OpenSchema(
                          adSchemaColumns,
                          VarArrayOf([Null, Null, TableNodeObject.Name]),
                          EmptyParam);

            while not Fields.EOF do begin
                // http://msdn.microsoft.com/en-us/library/ee266174(v=bts.10).aspx
                // http://msdn.microsoft.com/en-US/library/ms254969(v=VS.80).aspx
                // http://www.w3schools.com/ado/ado_datatypes.asp
                FieldName := VarToStr(Fields.Fields['COLUMN_NAME'].Value);
                FieldNodeObject := TFieldNodeObject.Create(FieldName);
                FieldNodeObject.Flags     := StrToIntDef(VarToStr(Fields.Fields['COLUMN_FLAGS'].Value), 0);
                FieldNodeObject.DataType  := Fields.Fields['DATA_TYPE'].Value;
                FieldNodeObject.Precision := StrToIntDef(VarToStr(Fields.Fields['NUMERIC_PRECISION'].Value), 0);
                FieldNodeObject.Scale     := StrToIntDef(VarToStr(Fields.Fields['NUMERIC_SCALE'].Value), 0);
                FieldNodeObject.FieldLen  := StrToIntDef(VarToStr(Fields.Fields['CHARACTER_MAXIMUM_LENGTH'].Value), 0);
                FieldNodeObject.FieldType := ADOTypeToFieldType(
                                                 FieldNodeObject.DataType,
                                                 FieldNodeObject.Precision,
                                                 FieldNodeObject.Scale,
                                                 TRUE);
                TreeView1.Items.AddChildObject(Node,
                                               FieldNodeObject.ToString,
                                               FieldNodeObject);
                Fields.MoveNext;
            end;
            Node.Expand(FALSE);
        end;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TAdoSchemaForm.TreeView1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
    FMouseX := X;
    FMouseY := Y;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

{ TFieldNodeObject }

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function TFieldNodeObject.ToString: String;
var
    TypeName : String;
begin
    TypeName := ADOTypeToStr(DataType);
    if (Flags and Ord(DBCOLUMNFLAGS_ISFIXEDLENGTH)) = 0 then begin
        if DataType = adChar then
            TypeName := ADOTypeToStr(adVarChar);
    end;
    if DataType in [adDecimal, adNumeric, adVarNumeric] then
        TypeName := TypeName + '(' + IntToStr(Precision) + ',' +
                                     IntToStr(Scale) + ')'
    else if FieldLen = $7FFFFFFF then
        TypeName := TypeName + '(max)'
    else if FieldLen > 0 then
        TypeName := TypeName + '(' + IntToStr(FieldLen) + ')';

    if (Flags and Ord(DBCOLUMNFLAGS_ISNULLABLE)) = 0 then
        TypeName := TypeName + ' not';

    TypeName := TypeName + ' NULL';

(*
    TypeName := GetEnumName(TypeInfo(TFieldType), Ord(FieldType));
    if FieldType in [ftFMTBcd, ftBcd] then
        TypeName := TypeName + '(' + IntToStr(Precision) + ',' +
                                     IntToStr(Scale) + ')';
*)
    Result := Name + '  ' + TypeName;
(*              + '  ' +
              IntToStr(FieldLen) + '  ' +
              IntToStr(Precision) + '  ' +
              IntToStr(Scale) + '  ' +
              IntToHex(Flags, 5); *)
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

end.
